import numpy as np
import itertools
from scipy.stats import bernoulli

class ContextualBandit():
    def __init__(self,
                 T,
                 n_arms,
                 n_features,
                 h,
                 trunc,
                 bits_unif=1,
                 noise_std=1.0,
                 quant_type=None,
                ):
        # number of rounds
        self.T = T
        # number of arms
        self.n_arms = n_arms
        # number of features for each arm
        self.n_features = n_features
        # average reward function
        # h : R^d -> R
        self.h = h

        self.bits_unif = bits_unif

        # standard deviation of Gaussian reward noise
        self.noise_std = noise_std

        self.quant_type = quant_type
        self.trunc = trunc

        # generate random features
        self.reset()


    @property
    def arms(self):
        """Return [0, ...,n_arms-1]
        """
        return range(self.n_arms)
        
    def reset(self):
        """Generate new features and new rewards.
        """
        self.reset_features()
        self.reset_rewards()

    def reset_features(self):
        """Generate normalized random N(0,1) features.
        """
        x = np.random.randn(self.T, self.n_arms, self.n_features)
        x /= np.repeat(np.linalg.norm(x, axis=-1, ord=2), self.n_features).reshape(self.T, self.n_arms, self.n_features)
        self.features = x

    def reset_rewards(self):
        """Generate rewards for each arm and each round,
        following the reward function h + Gaussian noise.
        """
        self.rewards = np.array(
            [
                self.h(self.features[t, k]) + self.noise_std*np.random.randn()\
                for t,k in itertools.product(range(self.T), range(self.n_arms))
            ]
        ).reshape(self.T, self.n_arms)

        if self.quant_type == 'our':
            norm_r = self.rewards/self.noise_std
            norm_f = np.floor(norm_r)
            p = norm_r-norm_f
            q_val = (norm_f + bernoulli.rvs(p, size=p.shape))*self.noise_std
            self.q_rewards = q_val
        elif self.quant_type == 'uniform':
            temp = np.clip(self.rewards,-1*self.trunc,self.trunc)
            scale = (2*self.trunc)/(2**self.bits_unif-1)
            norm_r = (temp+self.trunc)/scale
            norm_f = np.floor(norm_r)
            p = norm_r-norm_f
            q_val = (norm_f + bernoulli.rvs(p, size=p.shape))*scale
            self.q_rewards = q_val-self.trunc
        else:
            self.q_rewards = self.rewards

        # to be used only to compute regret, NOT by the algorithm itself
        means = np.array(
            [
                self.h(self.features[t, k])\
                for t,k in itertools.product(range(self.T), range(self.n_arms))
            ]
        ).reshape(self.T, self.n_arms)
        self.best_rewards_oracle = np.max(means, axis=1)
        self.best_actions_oracle = np.argmax(means, axis=1)
